import numpy as np
from skimage import segmentation, morphology, measure, color, feature, filters
from skimage import img_as_float
from scipy import ndimage
from sklearn.ensemble import RandomForestClassifier
from sklearn.neighbors import KNeighborsClassifier

# ------------------- Modification of segmentation ----------------------

def _split_labels(labels, mask, img, erosion_width=10, compactness=0):
    """
    Divide already labeled array ``labels`` according to array of annotations
    ``mask``.

    Parameters
    ----------

    labels : array of ints
        Array of labels.
    mask : array of ints
        Array with annotations.
    img : array, default None
        Image used for the segmentation.
    erosion_width : int, optional
        Number of pixels to erode on both sides of annotations. Decrease if
        boundaries should correspond to the exact annotations, increase if
        boundaries should rather follow strong image gradients.
    compactness : float, optional (default 0)
        Parameter of the compact waterhsed algorithm
        ``skimage.segmentation.watershed``, should be zero (for normal
        watershed) or greater than zero (for compact watershed).
    """
    out = np.copy(labels)
    bounding_boxes = ndimage.find_objects(labels)
    max_label = labels.max()
    annot_indices = np.unique(mask)[1:]
    count = max_label + 1
    shift = 0 if labels.min() > 0 else 1
    for annot_index in annot_indices:
        obj_label = (np.argmax(np.bincount(labels[mask == annot_index])[shift:]
                                )  + shift)
        # Get subarrays
        box = bounding_boxes[obj_label - 1]
        img_box = img[box]
        labels_box = labels[box]
        # Prepare watershed
        gradient_img = - ndimage.gaussian_gradient_magnitude(img_box, 2)
        mask_box = np.ones(img_box.shape, dtype=np.uint8)
        mask_box[mask[box] == annot_index] = 0
        mask_box = morphology.binary_erosion(mask_box,
                                             morphology.disk(erosion_width))
        masked_region = labels_box == obj_label
        mask_box[np.logical_not(masked_region)] = 0
        mask_box = measure.label(mask_box)
        res = segmentation.watershed(gradient_img, mask_box,
                                     compactness=compactness,
                                     mask=masked_region)
        out[box][res == 1] = count  # only modify one of the regions
        count += 1
    return out


def _merge_labels(labels, mask, skip_zero=True):
    """
    Merge objects covered by the same annotation, given an array of 
    labels defining objects and a labeled mask of annotations.

    Parameters
    ----------

    labels : array of ints
        Array of labels.
    mask : array of ints
        Array with annotations.

    """
    annot_indices = np.unique(mask)[1:]
    label_indices = np.arange(labels.max() + 1, dtype=np.int)
    for index in annot_indices:
        object_indices = np.unique(labels[mask == index])
        if skip_zero:
            object_indices = np.setdiff1d(object_indices, [0])
        label_indices[object_indices] = object_indices[0]
    relabeled, _, _ = segmentation.relabel_sequential(label_indices)
    return relabeled[labels]


def modify_segmentation(labels, mask, img=None, mode='split'):
    """
    Modify already labeled image according to annotations. In 'split' mode,
    each annotation is used to divide a label in two subregions. In 'merge'
    mode, objects covered by the same annotation are merged together.

    Parameters
    ----------

    labels : array of ints
        Array of labels.
    mask : binary of int array
        Array with annotations.
    img : array, default None
        Image used for the segmentation.
    mode: string, 'split' or 'merge'

    Returns
    -------

    out : array of ints
        New labels.
    """
    labels = np.asarray(labels)
    mask = measure.label(mask)
    if img is None:
        img = np.zeros_like(mask)
    
    if mode == 'split':
        return _split_labels(labels, mask, img)
    elif mode == 'merge':
        return _merge_labels(labels, mask)
    else:
        raise ValueError('mode should be either split or merge')


# ------------- Segmentation from markers ----------------------------

def watershed_segmentation(img, mask, sigma=4):
    """
    Watershed segmentation of image using annotations as label markers.

    The image used for the watershed is minus the gradient of the original
    image, convoluted with a Gaussian for more robustness.

    Parameters
    ----------

    img : ndarray
        image to be segmented
    mask : ndarray of ints
        binary array, each connected component corresponds to a different
        object to be segmented
    sigma : float
        standard deviation of Gaussian convoluting the gradient. Increase
        for smoother boundaries.
    """
    if img.ndim > 2:
        img = color.rgb2gray(img)
    labels = measure.label(mask)
    gradient_img = - ndimage.gaussian_gradient_magnitude(img, sigma)
    output = segmentation.watershed(gradient_img, labels)
    return output


def random_walker_segmentation(img, mask, beta=5000):
    """
    Random-walker segmentation of image using annotations as label markers.

    Parameters
    ----------

    img : ndarray
        image to be segmented
    mask : ndarray of ints
        binary array, each connected component corresponds to a different
        object to be segmented
    beta : float
        beta parameter of random walker, smaller beta results in more compact
        objects, larger beta in boundaries following gradients more
        accurately.

    Returns
    -------

    output : ndarray of ints
        image of labels

    Notes
    -----

    see skimage.segmentation.random_walker
    """
    if img.ndim > 2:
        img = color.rgb2gray(img)
    labels = measure.label(mask)
    output = segmentation.random_walker(img, labels, beta=beta, mode='cg_mg')
    return output


def _compute_features_gabor(im):
    gabor_frequencies = np.logspace(-5, 0.5, num=6, base=2)
    thetas = [0, np.pi/4., np.pi/2]
    nb_fq = len(gabor_frequencies) * len(thetas)
    im = np.atleast_3d(im)
    im_gabor = np.empty((im.shape[-1], nb_fq) + im.shape[:2])
    for ch in range(im.shape[-1]):
        img = img_as_float(im[..., ch])
        for i_fq, fq in enumerate(gabor_frequencies):
            for i_th, theta in enumerate(thetas):
                tmp = filters.gabor(img, fq, theta=theta)
                im_gabor[ch, len(thetas) * i_fq + i_th] = \
                                    np.abs(tmp[0] + 1j * tmp[1])
    return im_gabor


def random_forest_segmentation(img, mask, mode='gabor'):
    """
    Segmentation of image using machine learning. Annotations are used 
    as training set for a random forest classifier. Features used for the
    classification are Gabor or Daisy features.

    Parameters
    ----------

    img : ndarray
        Image to be segmented
    mask : ndarray of ints
        Binary array, each connected component corresponds to a different
        object to be segmented
    mode : string
        Type of features used in classification, can be 'daisy' or 'gabor'.
    """
    labels = measure.label(mask)
    if mode == 'daisy':
        if img.ndim > 2:
            img = color.rgb2gray(img)
        radius = 15
        features = feature.daisy(img, step=1, radius=radius, rings=2,
                                 histograms=4)
        crop_labels = labels[radius:-radius, radius:-radius]
    elif mode == 'gabor':
        features = _compute_features_gabor(img)
        nb_ch, nb_fq, sh_1, sh_2 = features.shape
        features = features.reshape((nb_ch * nb_fq, sh_1, sh_2))
        features = np.moveaxis(features, 0, -1)
        crop_labels = labels
    X_train = features[crop_labels > 0, :]
    Y_train = crop_labels[crop_labels > 0]
    rf = RandomForestClassifier(n_estimators=100)
    rf.fit(X_train, Y_train)
    output = rf.predict(features.reshape(-1, features.shape[2]))
    output = output.reshape(crop_labels.shape)
    output[crop_labels > 0] = crop_labels[crop_labels > 0]
    return output


def segmentation_generic(img, mask, mode='watershed'):
    """
    Segmentation of image using annotations as markers (seeds), using
    different algorithms.

    Parameters
    ----------

    img : ndarray
        Image to be segmented
    mask : ndarray of ints
        Binary array, each connected component corresponds to a different
        object to be segmented
    mode : 'string', optional
        Algorithm to be used, can be 'watershed', 'random_walker' or
        'random_forest'.
    """
    if mode=='watershed':
        return watershed_segmentation(img, mask)
    elif mode=='random_walker':
        return random_walker_segmentation(img, mask)
    elif mode=='random_forest':
        return random_forest_segmentation(img, mask)
    else:
        raise NotImplementedError


## -------------------- Background removal ----------------------------


def superpixel_color_segmentation(im, mask, mode='bbox', remove_holes='all',
        object='single', keep_label=True):
    if not mask.shape == im.shape:
        big_mask = np.zeros(im.shape[:2], dtype=np.uint8)
        big_mask[:mask.shape[0], :mask.shape[1]] = mask
        mask = big_mask
    if mask.max() == 1:
        mask = annotation_to_background_mask(mask, mode=mode)
    px = segmentation.felzenszwalb(im, scale=4)
    colors = np.stack([ndimage.mean(im[..., i], px,
                       index=np.arange(0, px.max() + 1))
                       for i in range(3)]).T
    nb_phases = len(np.unique(mask)) - 1
    indices = [np.unique(px[mask == i]) for i in range(1, nb_phases + 1)]
    target = np.concatenate([(i + 1) * np.ones(len(indices[i]))
                             for i in range(len(indices))])
    training_set = np.vstack([colors[indices_i] for indices_i in indices])
    clf = KNeighborsClassifier()
    clf.fit(training_set, target)
    res = clf.predict(colors)
    if keep_label:
        for i in range(1, nb_phases + 1):
            res[indices[i - 1]] = i
    res_img = res[px]
    mask_res = res_img == 1
    if object == 'single':
        labels = measure.label(mask_res)
        index = np.argmax(np.bincount(labels.ravel())[1:]) + 1
        mask_res = labels == index

    if remove_holes == 'all':
        return ndimage.binary_fill_holes(mask_res)
    elif np.isscalar(remove_holes):
        return morphology.remove_small_holes(mask_res, remove_holes)


def annotation_to_background_mask(mask_annotation, mode='bbox'):
    mask = np.copy(mask_annotation).astype(np.uint8)
    if mode == 'bbox':
        sls = ndimage.find_objects(mask_annotation)[0]
        mask[:int(0.9 * sls[0].start)] = 2
        mask[int(1.1 * sls[0].stop):] = 2
        mask[:, :int(0.9 * sls[1].start)] = 2
        mask[:, int(1.1 * sls[1].stop):] = 2
    if mode == 'cvxhull':
        cvx = morphology.convex_hull_image(mask_annotation)
        cvx = morphology.binary_dilation(cvx, np.ones((11, 11)))
        mask[np.logical_not(cvx)] = 2
    return mask



import numpy as np
import json
from skimage import draw, morphology
from scipy import ndimage


def _indices_of_path(path, scale=1):
    """
    Retrieve pixel indices (integer values). 

    Parameters
    ----------

    path: SVG-like path formatted as JSON string
        The path is formatted like
        ['M', x0, y0],
        ['Q', xc1, yc1, xe1, ye1],
        ['Q', xc2, yc2, xe2, ye2],
        ...
        ['L', xn, yn]
        where (xc, yc) are for control points and (xe, ye) for end points.

    Notes
    -----

    I took a weight of 1 and it seems fine from visual inspection.
    """
    rr, cc = [], []
    for (Q1, Q2) in zip(path[:-2], path[1:-1]):
        inds = draw.bezier_curve(round(Q1[-1] / scale), round(Q1[-2] / scale), 
                                 round(Q2[2] / scale), round(Q2[1] / scale), 
                                 round(Q2[4] / scale), round(Q2[3] / scale), 1)
        rr += list(inds[0])
        cc += list(inds[1])
    return rr, cc


def parse_jsonstring(string, shape=None, scale=1):
    """
    Parse JSON string to draw the path saved by react-sketch.

    Up to now only path objects are processed (created with Pencil tool).

    Parameters
    ----------

    data : str
        JSON string of data
    shape: tuple, optional
        shape of returned image.

    Returns
    -------

    mask: ndarray of bools
        binary array where the painted regions are one.
    """
    if shape is None:
        shape = (500, 500)
    mask = np.zeros(shape, dtype=np.bool)
    try:
        data = json.loads(string)
    except:
        return mask
    scale = 1
    for obj in data['objects']:
        if obj['type'] == 'image':
            scale = obj['scaleX']
        elif obj['type'] == 'path':
            scale_obj = obj['scaleX']
            inds = _indices_of_path(obj['path'], scale=scale / scale_obj)
            radius = round(obj['strokeWidth'] / 2. / scale)
            mask_tmp = np.zeros(shape, dtype=np.bool)
            mask_tmp[inds[0], inds[1]] = 1
            mask_tmp = ndimage.binary_dilation(mask_tmp,
                                                  morphology.disk(radius))
            mask += mask_tmp
    return mask


def parse_jsonstring_line(string):
    """
    Return geometry of line objects.

    Parameters
    ----------

    data : str
        JSON string of data

    """
    try:
        data = json.loads(string)
    except:
        return None
    scale = 1
    props = []
    for obj in data['objects']:
        if obj['type'] == 'image':
            scale = obj['scaleX']
        elif obj['type'] == 'line':
            length = np.sqrt(obj['width']**2 + obj['height']**2)
            scale_factor = obj['scaleX'] / scale
            props.append([scale_factor * length,
                          scale_factor * obj['width'],
                          scale_factor * obj['height'],
                          scale_factor * obj['left'],
                          scale_factor * obj['top']])
    return (np.array(props)).astype(np.int)


def parse_jsonfile(filename, shape=None):
    with open(filename, 'r') as fp:
        string = json.load(fp)
    return parse_jsonstring(string, shape=shape)





from dash_canvas.utils import parse_jsonfile


def test_parse_jsonfile():
    shape = (433, 640)
    mask = parse_jsonfile('tests/data_test.json', shape=shape)
    assert mask.sum() > 0
    assert mask.shape == shape



from dash_canvas.utils import watershed_segmentation, modify_segmentation
from skimage import data, segmentation, morphology, measure
import numpy as np
from scipy import ndimage


def test_watershed_segmentation():
    img = np.zeros((20, 20))
    img[2:6, 2:6] = 1
    img[10:15, 10:15] = 1
    mask = np.zeros_like(img, dtype=np.uint8)
    mask[4, 4] = 1
    mask[12, 12] = 2
    res = watershed_segmentation(img, mask, sigma=0.1)
    assert np.all(res[2:6, 2:6] == 1)
    assert np.all(res[10:15, 10:15] == 2)


def test_split_segmentation():
    img = np.zeros((100, 100), dtype=np.uint8)
    img[:40, 55:] = 1
    img[40:, :30] = 2
    img[40:, 30:65] = 3
    img[40:, 65:] = 4 
    img = ndimage.rotate(img, 20)
    img = measure.label(img)
    img = morphology.remove_small_objects(img, 20)
    img = segmentation.relabel_sequential(img)[0]

    mask = np.zeros_like(img)
    mask[2:53, 75] = 1
    mask[100, 17:60] = 1

    # Labels start at 1
    seg = modify_segmentation(img, measure.label(mask), mode='split')
    assert len(np.unique(seg)) == len(np.unique(img)) + 2

    # Labels start at 0
    seg = modify_segmentation(img + 1, measure.label(mask))
    assert len(np.unique(seg)) == len(np.unique(img)) + 2


def test_merge_segmentation():
    img = np.zeros((20, 20), dtype=np.uint8)
    img[:10, :10] = 1
    img[10:, :10] = 2
    mask = np.zeros_like(img)
    mask[:, 5] = 1
    seg = modify_segmentation(img, mask, mode='merge')
    assert np.all(np.unique(seg) == np.array([0, 1]))



import numpy as np

def contrast_adjust(img, intensity):
    print("contrast adjust", intensity)
    contrast_range = 255 ** (2 - 2*intensity)
    img_f = img.astype(np.float)
    img_f = np.clip((img_f - 127) * 255. / contrast_range + 127, 0, 255)
    return img_f.astype(np.uint8)


def brightness_adjust(img, intensity):
    offset = int((intensity - 0.5) * 255)
    if intensity < 0.5:
        img = np.clip(img, np.abs(offset), 255)
        return (img + offset).astype(np.uint8)
    elif intensity > 0.5:
        img = np.clip(img, None, 255 - offset)
        return (img + offset).astype(np.uint8)
    else:
        return img




import dash_core_components as dcc
import dash_html_components as html


def image_upload_zone(name, multiple=False, width=100):
    return dcc.Upload(
                id=name,
                children=[
                    'Drag and Drop or ',
                    html.A('Select an Image')],
                style={'width': str(width) + '%',
                       'height': '50px',
                       'lineHeight': '50px',
                       'borderWidth': '1px',
                       'borderStyle': 'dashed',
                       'borderRadius': '5px',
                       'textAlign': 'center'
                       },
                accept='image/*',
                multiple=multiple,
                )




# AUTO GENERATED FILE - DO NOT EDIT

from dash.development.base_component import Component, _explicitize_args


class AlternativeCanvas(Component):
    """A AlternativeCanvas component.
AlternativeCanvas is an example component.
It takes a property, `label`, and
displays it.
It renders an input with the property `value`
which is editable by the user.

Keyword arguments:
- id (string; optional): The ID used to identify this component in Dash callbacks
- label (string; required): A label that will be printed when this component is rendered.
- value (string; optional): The value displayed in the input
- lineColor (string; optional): the color of the line

Available events: """
    @_explicitize_args
    def __init__(self, id=Component.UNDEFINED, label=Component.REQUIRED, value=Component.UNDEFINED, lineColor=Component.UNDEFINED, **kwargs):
        self._prop_names = ['id', 'label', 'value', 'lineColor']
        self._type = 'AlternativeCanvas'
        self._namespace = 'dash_canvas'
        self._valid_wildcard_attributes =            []
        self.available_events = []
        self.available_properties = ['id', 'label', 'value', 'lineColor']
        self.available_wildcard_properties =            []

        _explicit_args = kwargs.pop('_explicit_args')
        _locals = locals()
        _locals.update(kwargs)  # For wildcard attrs
        args = {k: _locals[k] for k in _explicit_args if k != 'children'}

        for k in ['label']:
            if k not in args:
                raise TypeError(
                    'Required argument `' + k + '` was not specified.')
        super(AlternativeCanvas, self).__init__(**args)

    def __repr__(self):
        if(any(getattr(self, c, None) is not None
               for c in self._prop_names
               if c is not self._prop_names[0])
           or any(getattr(self, c, None) is not None
                  for c in self.__dict__.keys()
                  if any(c.startswith(wc_attr)
                  for wc_attr in self._valid_wildcard_attributes))):
            props_string = ', '.join([c+'='+repr(getattr(self, c, None))
                                      for c in self._prop_names
                                      if getattr(self, c, None) is not None])
            wilds_string = ', '.join([c+'='+repr(getattr(self, c, None))
                                      for c in self.__dict__.keys()
                                      if any([c.startswith(wc_attr)
                                      for wc_attr in
                                      self._valid_wildcard_attributes])])
            return ('AlternativeCanvas(' + props_string +
                   (', ' + wilds_string if wilds_string != '' else '') + ')')
        else:
            return (
                'AlternativeCanvas(' +
                repr(getattr(self, self._prop_names[0], None)) + ')')



import numpy as np
from skimage import io, measure, feature
from scipy import ndimage


def autocrop(img):
    """
    Remove zero-valued rectangles at the border of the image.

    Parameters
    ----------

    img: ndarray
        Image to be cropped
    """
    slices = ndimage.find_objects(img > 0)[0]
    return img[slices]


def _blending_mask(shape):
    mask = np.zeros(shape, dtype=np.int)
    mask[1:-1, 1:-1] = 1
    return ndimage.distance_transform_cdt(mask) + 1


def register_tiles(imgs, n_rows, n_cols, overlap_global=None,
                   overlap_local=None, pad=None, blending=True):
    """
    Stitch together overlapping tiles of a mosaic, using Fourier-based
    registration to estimate the shifts between neighboring tiles.

    Parameters
    ----------

    imgs: array of tiles, of shape (n_rows, n_cols, l_r, l_r) with (l_c, l_r)
        the shape of individual tiles.
    n_rows: int
        number of rows of the mosaic.
    n_cols : int
        number of columns of the mosaic.
    overlap_global : float
        Fraction of overlap between tiles.
    overlap_local : dictionary
        Local overlaps between pairs of tiles. overlap_local[(i, j)] is a pair
        of (x, y) shifts giving the 2D shift vector between tiles i and j.
        Indices (i, j) are the raveled indices of the tile numbers.
    pad : int
        Value of the padding used at the border of the stitched image. An
        autocrop is performed at the end to remove the unnecessary padding.

    Notes
    -----

    Fourier-based registration is used in this function
    (skimage.feature.register_translation).
    """
    if pad is None:
        pad = 200
    l_r, l_c = imgs.shape[2:4]
    if overlap_global is None:
        overlap_global = 0.15
    overlap_value = int(float(overlap_global) * l_r)
    imgs = imgs.astype(np.float)
    if blending:
        blending_mask = _blending_mask((l_r, l_c))
    else:
        blending_mask = np.ones((l_r, l_c))

    if imgs.ndim == 4:
        canvas = np.zeros((2 * pad + n_rows * l_r, 2 * pad + n_cols * l_c),
                      dtype=imgs.dtype)
    else:
        canvas = np.zeros((2 * pad + n_rows * l_r, 2 * pad + n_cols * l_c, 3),
                      dtype=imgs.dtype)
        blending_mask = np.dstack((blending_mask, )*3)
    weights = np.zeros_like(canvas)
    init_r, init_c = pad, pad
    weighted_img = imgs[0, 0] * blending_mask
    canvas[init_r:init_r + l_r, init_c:init_c + l_c] = weighted_img
    weights[init_r:init_r + l_r, init_c:init_c + l_c] = blending_mask
    shifts = np.empty((n_rows, n_cols, 2), dtype=np.int)
    shifts[0, 0] = init_r, init_c

    for i_rows in range(n_rows):
        # Shifts between rows
        if i_rows >= 1:
            index_target = np.ravel_multi_index((i_rows, 0), (n_rows, n_cols))
            index_orig = index_target - n_cols
            try:
                overlap = overlap_local[(index_orig, index_target)]
            except (KeyError, TypeError):
                overlap = np.array([overlap_value, 0])
            init_r, init_c = shifts[i_rows - 1, 0]
            init_r += l_r
            shift_vert = feature.register_translation(
                    imgs[i_rows - 1, 0, -overlap[0]:, :(l_c - overlap[1])],
                    imgs[i_rows, 0, :overlap[0], -(l_c - overlap[1]):])[0]
            init_r += int(shift_vert[0])  - overlap[0]
            init_c += int(shift_vert[1]) - overlap[1]
            shifts[i_rows, 0] = init_r, init_c
            # Fill canvas and weights
            weighted_img = imgs[i_rows, 0] * blending_mask
            canvas[init_r:init_r + l_r, init_c:init_c + l_c] += weighted_img
            weights[init_r:init_r + l_r, init_c:init_c + l_c] += blending_mask
        # Shifts between columns
        for j_cols in range(n_cols - 1):
            index_orig = np.ravel_multi_index((i_rows, j_cols),
                                              (n_rows, n_cols))
            index_target = index_orig + 1
            try:
                overlap = overlap_local[(index_orig, index_target)]
            except (KeyError, TypeError):
                overlap = np.array([0, overlap_value])
            init_c += l_c
            shift_horiz = feature.register_translation(
                imgs[i_rows, j_cols, - (l_r - overlap[0]):, -overlap[1]:],
                imgs[i_rows, j_cols + 1, : l_r - overlap[0], :overlap[1]])[0]
            init_r += int(shift_horiz[0]) + overlap[0]
            init_c += int(shift_horiz[1]) - overlap[1]
            shifts[i_rows, j_cols + 1] = init_r, init_c
            # Fill canvas and weights
            weighted_img = imgs[i_rows, j_cols + 1] * blending_mask
            canvas[init_r:init_r + l_r, init_c:init_c + l_c] += weighted_img
            weights[init_r:init_r + l_r, init_c:init_c + l_c] += blending_mask

    canvas /= (weights + 1.e-5)
    return autocrop(np.rint(canvas).astype(np.uint8))



from .image_processing_utils import (watershed_segmentation,
                                     random_walker_segmentation,
                                     random_forest_segmentation,
                                     segmentation_generic,
                                     superpixel_color_segmentation,
                                     modify_segmentation)
from .registration import register_tiles, autocrop
from .parse_json import parse_jsonstring, parse_jsonstring_line, parse_jsonfile
from .io_utils import array_to_data_url, image_string_to_PILImage
from .plot_utils import image_with_contour
from .exposure import brightness_adjust, contrast_adjust

__all__ = ['array_to_data_url',
           'autocrop',
           'brightness_adjust',
           'contrast_adjust',
           'image_string_to_PILImage',
           'image_with_contour',
           'modify_segmentation',
           'parse_jsonfile',
           'parse_jsonstring',
           'parse_jsonstring_line',
           'random_forest_segmentation',
           'random_walker_segmentation',
           'register_tiles',
           'segmentation_generic',
           'superpixel_color_segmentation',
           'watershed_segmentation']




import plotly.graph_objs as go
import PIL
import numpy as np
from skimage import color, img_as_ubyte


def image_with_contour(img, labels, shape=None):
    """
    Figure with contour plot of labels superimposed on background image.

    Parameters
    ----------

    img : URL, dataURI or ndarray
        Background image. If a numpy array, it is transformed into a PIL
        Image object.
    labels : 2D ndarray
        Contours are the isolines of labels.
    shape: tuple, optional
        Shape of the arrays, to be provided if ``img`` is not a numpy array.
    """
    try:
        sh_y, sh_x = shape if shape is not None else img.shape
    except AttributeError:
        print('''the shape of the image must be provided with the
                 ``shape`` parameter if ``img`` is not a numpy array''')
    if type(img) == np.ndarray:
        img = img_as_ubyte(color.gray2rgb(img))
        img = PIL.Image.fromarray(img)
    labels = labels.astype(np.float)
    # Contour plot of segmentation
    cont = go.Contour(z=labels[::-1],
            contours=dict(coloring='lines'),
            line=dict(width=3),
            showscale=False,
            )
    # Layout
    layout= go.Layout(
            images = [dict(
                  source=img,
                  xref="x",
                  yref="y",
                  x=0,
                  y=sh_y,
                  sizex=sh_x,
                  sizey=sh_y,
                  sizing="contain",
                  layer="below")],
            xaxis=dict(
                  showgrid=False,
                  zeroline=False,
                  showline=False,
                  ticks='',
                  showticklabels=False,
                  ),
            yaxis=dict(
                  showgrid=False,
                  zeroline=False,
                  showline=False,
                  scaleanchor="x",
                  ticks='',
                  showticklabels=False,
    ))
    fig = go.Figure(data=[cont], layout=layout)
    return fig


if __name__ == '__main__':
    from skimage import data
    import plotly.plotly as py
    camera = data.camera()
    fig = image_with_contour(camera, camera > 150)
    py.iplot(fig)



import shutil
import sys


from pytest_dash.wait_for import (
    wait_for_text_to_equal,
    wait_for_element_by_css_selector
)
from pytest_dash.application_runners import import_app

def test_install(dash_threaded):

    # Add the generated project to the path so it can be loaded from usage.py
    # It lies somewhere in a temp directory created by pytest-cookies
    sys.path.insert(0, '..')

    # Test that `usage.py` works after building the default component.
    dash_threaded(import_app('app'))




from dash_canvas.utils import register_tiles
import numpy as np
from skimage import data, color
import matplotlib.pyplot as plt

def test_stitching_one_row():
    im = data.moon()
    l = 256

    n_rows = 1
    n_cols = im.shape[1] // l

    init_i, init_j = 0, 0

    imgs = np.empty((n_rows, n_cols, l, l))

    overlap_h = [5, 50]

    i = 0
    for j in range(n_cols):
        sub_im = im[init_i:init_i + l, init_j:init_j + l]
        imgs[i, j] = sub_im
        init_j += l - overlap_h[1]

    stitch = register_tiles(imgs, n_rows, n_cols, overlap_global=0.2, pad=l//2)
    delta = im[:l, :stitch.shape[1]].astype(np.float) - stitch.astype(np.float)
    assert np.all(delta == 0)
    stitch = register_tiles(imgs, n_rows, n_cols, overlap_global=0.2, pad=l//2,
                            blending=False)
    delta = im[:l, :stitch.shape[1]].astype(np.float) - stitch.astype(np.float)
    assert np.all(delta == 0)

    # local_overlap
    stitch = register_tiles(imgs, n_rows, n_cols, overlap_global=0.5, pad=l//2,
                            overlap_local={(0, 1):[0, 45]})
    delta = im[:l, :stitch.shape[1]].astype(np.float) - stitch.astype(np.float)
    assert np.all(delta == 0)


def test_stitching_two_rows():
    im = data.moon()
    l = 256
    # two rows
    n_rows = 2
    n_cols = im.shape[1] // l
    init_i, init_j = 0, 0
    overlap_h = [5, 50]
    overlap_v = 40

    imgs = np.empty((n_rows, n_cols, l, l))
    for i in range(n_rows):
        for j in range(n_cols):
            sub_im = im[init_i:init_i + l, init_j:init_j + l]
            imgs[i, j] = sub_im
            init_j += l - overlap_h[1]
        init_j = 0
        init_i += l - overlap_v

    stitch = register_tiles(imgs, n_rows, n_cols, overlap_global=0.2, pad=l//2)
    delta = im[:stitch.shape[0], :stitch.shape[1]].astype(np.float) - stitch.astype(np.float)
    print(delta.mean())
    assert np.all(delta == 0)


def test_stitching_color():
    im = color.gray2rgb(data.moon())
    l = 256

    n_rows = 1
    n_cols = im.shape[1] // l

    init_i, init_j = 0, 0

    imgs = np.empty((n_rows, n_cols, l, l, 3))

    overlap_h = [5, 50]

    i = 0
    for j in range(n_cols):
        sub_im = im[init_i:init_i + l, init_j:init_j + l]
        imgs[i, j] = sub_im
        init_j += l - overlap_h[1]

    stitch = register_tiles(imgs, n_rows, n_cols, overlap_global=0.2, pad=l//2)
    delta = im[:l, :stitch.shape[1]].astype(np.float) - stitch.astype(np.float)
    assert np.all(delta == 0)




import json
import os
import setuptools
from setuptools import setup


with open(os.path.join('dash_canvas', 'package.json')) as f:
    package = json.load(f)

package_name = package["name"].replace(" ", "_").replace("-", "_")


setup(
    name=package_name,
    version=package["version"],
    author=package['author'],
    url='https://github.com/plotly/dash-canvas',
    packages=setuptools.find_packages(),
    include_package_data=True,
    license=package['license'],
    description=package['description'] if 'description' in package else package_name,
    install_requires=['dash', 'dash-html-components', 'scikit-image', 'Pillow']
)



# AUTO GENERATED FILE - DO NOT EDIT

from dash.development.base_component import Component, _explicitize_args


class DashCanvas(Component):
    """A DashCanvas component.
Canvas component for drawing on a background image and selecting
regions.

Keyword arguments:
- id (string; optional): The ID used to identify this component in Dash callbacks
- label (string; optional): A label that will be printed when this component is rendered.
- image_content (string; optional): Image data string, formatted as png or jpg data string. Can be
generated by utils.io_utils.array_to_data_string.
- zoom (number; optional): Zoom factor
- width (number; optional): Width of the canvas
- height (number; optional): Height of the canvas
- scale (number; optional): Scaling ratio between canvas width and image width
- tool (string; optional): Selection of drawing tool, among ["pencil", "pan", "circle",
"select", "line"].
- lineWidth (number; optional): Width of drawing line (in pencil mode)
- lineColor (string; optional): Color of drawing line (in pencil mode). Can be a text string,
like 'yellow', 'red', or a color triplet like 'rgb(255, 0, 0)'.
Alpha is possible with 'rgba(255, 0, 0, 0.5)'.
- goButtonTitle (string; optional): Title of button
- filename (string; optional): Name of image file to load (URL string)
- trigger (number; optional): Counter of how many times the save button was pressed
(to be used mostly as input)
- json_data (string; optional): Sketch content as JSON string, containing background image and
annotations. Use utils.parse_json.parse_jsonstring to parse
this string.
- hide_buttons (list; optional): Names of buttons to hide. Names are "zoom", "pan", "line", "pencil",
"undo", "select".

Available events: """
    @_explicitize_args
    def __init__(self, id=Component.UNDEFINED, label=Component.UNDEFINED, image_content=Component.UNDEFINED, zoom=Component.UNDEFINED, width=Component.UNDEFINED, height=Component.UNDEFINED, scale=Component.UNDEFINED, tool=Component.UNDEFINED, lineWidth=Component.UNDEFINED, lineColor=Component.UNDEFINED, goButtonTitle=Component.UNDEFINED, filename=Component.UNDEFINED, trigger=Component.UNDEFINED, json_data=Component.UNDEFINED, hide_buttons=Component.UNDEFINED, **kwargs):
        self._prop_names = ['id', 'label', 'image_content', 'zoom', 'width', 'height', 'scale', 'tool', 'lineWidth', 'lineColor', 'goButtonTitle', 'filename', 'trigger', 'json_data', 'hide_buttons']
        self._type = 'DashCanvas'
        self._namespace = 'dash_canvas'
        self._valid_wildcard_attributes =            []
        self.available_events = []
        self.available_properties = ['id', 'label', 'image_content', 'zoom', 'width', 'height', 'scale', 'tool', 'lineWidth', 'lineColor', 'goButtonTitle', 'filename', 'trigger', 'json_data', 'hide_buttons']
        self.available_wildcard_properties =            []

        _explicit_args = kwargs.pop('_explicit_args')
        _locals = locals()
        _locals.update(kwargs)  # For wildcard attrs
        args = {k: _locals[k] for k in _explicit_args if k != 'children'}

        for k in []:
            if k not in args:
                raise TypeError(
                    'Required argument `' + k + '` was not specified.')
        super(DashCanvas, self).__init__(**args)

    def __repr__(self):
        if(any(getattr(self, c, None) is not None
               for c in self._prop_names
               if c is not self._prop_names[0])
           or any(getattr(self, c, None) is not None
                  for c in self.__dict__.keys()
                  if any(c.startswith(wc_attr)
                  for wc_attr in self._valid_wildcard_attributes))):
            props_string = ', '.join([c+'='+repr(getattr(self, c, None))
                                      for c in self._prop_names
                                      if getattr(self, c, None) is not None])
            wilds_string = ', '.join([c+'='+repr(getattr(self, c, None))
                                      for c in self.__dict__.keys()
                                      if any([c.startswith(wc_attr)
                                      for wc_attr in
                                      self._valid_wildcard_attributes])])
            return ('DashCanvas(' + props_string +
                   (', ' + wilds_string if wilds_string != '' else '') + ')')
        else:
            return (
                'DashCanvas(' +
                repr(getattr(self, self._prop_names[0], None)) + ')')



from PIL import Image
from io import BytesIO
import base64


def array_to_data_url(img, dtype=None):
    """
    Converts numpy array to data string, using Pillow.

    The returned image string has the right format for the ``image_content``
    property of DashCanvas.

    Parameters
    ==========

    img : numpy array

    Returns
    =======

    image_string: str
    """
    if dtype is not None:
        img = img.astype(dtype)
    pil_img = Image.fromarray(img)
    buff = BytesIO()
    pil_img.save(buff, format="png")
    prefix = b'data:image/png;base64,'
    image_string = (prefix + base64.b64encode(buff.getvalue())).decode("utf-8")
    return image_string


def image_string_to_PILImage(image_string):
    """
    Converts image string to PIL image object.
    """
    return Image.open(BytesIO(base64.b64decode(image_string[22:])))
